﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Data.SqlTypes;

namespace MariCam
{
    public class Program
    {
        public enum ConnectionType : int
        {
            CSI = 0,
            USB = 1
        }

        public enum FrameType : int
        {
            RAW = 0,
            AMPLITUDE = 1,
            DEPTH = 2
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct FrameFormat
        {
            public ushort Width;

            public ushort Height;

            public int FrameType;

            public ulong Timestamp;
        }

        public const string LibraryPath = "/usr/lib/libArducamDepthCamera2c.so";

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr createArducamDepthCamera();

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr init(IntPtr camera, ConnectionType connectionType, int deviceNumber);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr start(IntPtr camera, FrameType frameType);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static unsafe extern IntPtr requestFrame(IntPtr camera, int timeout);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern IntPtr releaseFrame(IntPtr camera, IntPtr frameBuffer);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern FrameFormat getFormat(IntPtr camera, int frameType);

        [System.Runtime.InteropServices.DllImport(LibraryPath)]
        public static extern IntPtr getDepthData(IntPtr frameBuffer);

        public static unsafe void Main(string[] args)
        {
            var camera = createArducamDepthCamera();

            init(camera, ConnectionType.CSI, 0);

            start(camera, FrameType.DEPTH);

            var tempBuffer = requestFrame(camera, 200);

            var tempFormat = getFormat(tempBuffer, 2);

            var buffer = new Bitmap(tempFormat.Width, tempFormat.Height, PixelFormat.Format32bppArgb);

            releaseFrame(camera, tempBuffer);

            var form = new Form();

            form.Text = "MariCam";

            form.WindowState = FormWindowState.Maximized;

            form.StartPosition = FormStartPosition.CenterScreen;

            form.Show();

            form.Paint += (object sender, PaintEventArgs e) =>
            {
                var frameBuffer = requestFrame(camera, 200);

                var format = getFormat(frameBuffer, 2);

                var graphics = e.Graphics;

                graphics.InterpolationMode = InterpolationMode.Bicubic;

                graphics.Clear(Color.Black);

                float* depthData = (float*)getDepthData(frameBuffer);

                float min = float.MaxValue;
                float max = float.MinValue;

                for (int y = 0; y < format.Height; y++)
                {
                    for (int x = 0; x < format.Width; x++)
                    {
                        var depth = *(depthData + (y * format.Width) + x);

                        if (depth < min)
                            min = depth;

                        if (depth > max)
                            max = depth;
                    }
                }

                float minNew = -1.0f;
                float maxNew = 1.0f;

                var bufferData = buffer.LockBits(new Rectangle(0, 0, buffer.Width, buffer.Height), ImageLockMode.ReadWrite, buffer.PixelFormat);

                var bufferPixels = (uint*)bufferData.Scan0;

                for (int y = 0; y < format.Height; y++)
                {
                    for (int x = 0; x < format.Width; x++)
                    {
                        //maxNew − minNew / maxOld − minOld * (v − minOld) + minNew
                        var intensity = ((float)(maxNew - minNew) / (float)(max - min) * ((*(depthData + (y * format.Width) + x)) - min) + minNew);

                        byte r = (intensity < 0.0) ? (byte)0 : (intensity >= 0.5) ? (byte)255 : (byte)(intensity / 0.5 * 255);
                        byte g = (intensity < -0.5) ? (byte)((intensity + 1) / 0.5 * 255) : (intensity > 0.5) ? (byte)(255 - ((intensity - 0.5) / 0.5 * 255)) : (byte)255;
                        byte b = (intensity > 0.0) ? (byte)0 : (intensity <= -0.5) ? (byte)255 : (byte)(intensity * -1.0 / 0.5 * 255);

                        *(bufferPixels + ((y * format.Width) + x)) = ((uint)255 << 24) | ((uint)(r) << 16) | ((uint)(g) << 8) | (byte)(b);
                    }
                }

                buffer.UnlockBits(bufferData);

                graphics.DrawImage(buffer, new Rectangle(0, 0, form.ClientSize.Width, form.ClientSize.Height));

                releaseFrame(camera, frameBuffer);

                form.Invalidate();
            };

            form.Invalidate();

            Application.DoEvents();
        }
    }
}
